#--
# *** This code is copyright 2004 by Gavin Kistner
# *** It is covered under the license viewable at http://phrogz.net/JS/_ReuseLicense.txt
# *** Reuse or modification is free provided you abide by the terms of that license.
# *** (Including the first two lines above in your source code usually satisfies the conditions.)
#++
# Common classes and extensions to make Ruby a little more pleasant to develop with.
# (Or, if you prefer, horrific hacks which destroy the purity of, and all that is Good about, Ruby.)
# Author::     Gavin Kistner  (mailto:!@phrogz.net)
# Copyright::  Copyright (c)2004 Gavin Kistner
# License::    See http://Phrogz.net/JS/_ReuseLicense.txt for details
# Version::    0.5
# Full Code::  link:../basiclibrary.rb


# Extensions to the Array class
class Array

  # Processes each unique pair (of indices, not value) in the array by yielding them to the supplied block.
  #
  #   a = [1,2,3,4]
  #   a.each_unique_pair{ |a,b|
  #     puts a+','+b
  #   }
  #   => 1,2
  #   => 1,3
  #   => 1,4
  #   => 2,3
  #   => 2,4
  #   => 3,4
  #
  # Does not guarantee the uniqueness of values. For example:
  #
  #   a = [1,2,1]
  #   a.each_unique_pair{ |a,b|
  #     puts a+','+b
  #   }
  #   => 1,2
  #   => 1,1
  #   => 2,1
  def each_unique_pair
    self.each_with_index{ |a,i|
      self[(i+1)..-1].each{ |b| yield a,b }
    }
  end

  # See Object#hier_inspect
  def hier_inspect(indent=0)
    t1 = "\t"*indent
    t2 = t1+"\t"
    s = "[\n"
    lastkey = self.length-1
    self.each_with_index{ |v,i|
      s << t2
      s << v.hier_inspect(indent+1)
      s << "," unless i==lastkey
      s << "\n"
    }
    s << t1+"]"
  end

end


# Extensions to the Hash class
class Hash

=begin
  # Causes Hash objects to behave like Object literals in JS, with their keys accessible via dot notation.
  #    foo = { 'name'=>'Gavin', 'age'=>31 }
  #    foo.weight=172
  #    print foo.name, foo.weight, foo.zzzz
  #    => 'Gavin172nil'
  def method_missing(meth,*args)
    if /=$/=~(meth=meth.id2name) then
      self[meth[0...-1]]= (args.length<2 ? args[0] : args)
    else
      self[meth]
    end
  end
=end

  # Converts all Symbol keys in the Hash to be String values, returning a new Hash.
  #   foo = { :name=>'Gavin', 'wife'=>:Lisa }
  #   foo.convert_symbol_keys  #=>  { "name"=>"Gavin", "wife"=>:Lisa }
  #   foo.inspect              #=>  { :name =>"Gavin", "wife"=>:Lisa }
  def convert_symbol_keys
    self.dup.convert_symbol_keys!
  end

  # Synonym for Hash#convert_symbol_keys, but modifies the receiver (and returns it).
  #   foo = { :name=>'Gavin', 'wife'=>:Lisa }
  #   foo.convert_symbol_keys!  #=>  { "name"=>"Gavin", "wife"=>:Lisa }
  #   foo.inspect               #=>  { "name"=>"Gavin", "wife"=>:Lisa }
  def convert_symbol_keys!
    self.each_key{ |k| self[k.id2name]=self.delete(k) if k.class==Symbol }
    self
  end


  # Converts all keys in the Hash to be Symbol values where possible, returning a new Hash.
  #   foo = { :name=>'Gavin', 'wife'=>:Lisa }
  #   foo.symbolify_keys  #=>  { :name=>"Gavin", :wife =>:Lisa }
  #   foo.inspect         #=>  { :name=>"Gavin", "wife"=>:Lisa }
  def symbolify_keys
    self.dup.symbolify_keys!
  end

  # Synonym for Hash#symbolify_keys, but modifies the receiver (and returns it).
  #   foo = { :name=>'Gavin', 'wife'=>:Lisa }
  #   foo.symbolify_keys!  #=>  { :name=>"Gavin", :wife=>:Lisa }
  #   foo.inspect          #=>  { :name=>"Gavin", :wife=>:Lisa }
  def symbolify_keys!
    self.each_key{ |k| self[k.to_sym]=self.delete(k) if k.respond_to?(:to_sym) }
    self
  end


  
  # Modifies the receiving Hash so that the value previously referred to by _oldkey_ is referenced by _newkey_; _oldkey_ is removed from the Hash.
  # If _oldkey_ does not exist as a key in the Hash, no change is effected.
  #
  # Returns a reference to the Hash.
  #   foo = { :name=>'Gavin', 'wife'=>:Lisa }
  #   foo.swapkey!('name',:name)     #=> { 'name'=>'Gavin', 'wife'=>:Lisa }
  #   foo.swapkey!('spouse','wife')  #=> { 'name'=>'Gavin', 'spouse'=>:Lisa }
  #   foo.swapkey!('bar','foo')      #=> { 'name'=>'Gavin', 'spouse'=>:Lisa }
  def swapkey!(newkey,oldkey)
    self[newkey]=self.delete(oldkey) if self.has_key?(oldkey)
    self
  end

  # Modifies the receiving Hash so that the value previously referred to by _oldkey_ is also referenced by _newkey_; _oldkey_ is retained in the Hash.
  # If _oldkey_ does not exist as a key in the Hash, no change is effected.
  #
  # Returns a reference to the Hash.
  #   foo = { :name=>'Gavin', 'wife'=>:Lisa }
  #   foo.alias!('name',:name)     => { :name=>'Gavin', 'name'=>'Gavin', 'wife'=>:Lisa }
  #   foo.alias!('spouse','wife')  => { :name=>'Gavin', 'name'=>'Gavin', 'wife'=>:Lisa, 'spouse'=>:Lisa }
  #   foo.alias!('bar','foo')      => { 'name'=>'Gavin', 'spouse'=>:Lisa }
  def alias!(newkey,oldkey)
    self[newkey]=self[oldkey] if self.has_key?(oldkey)
    self
  end

  # See Object#hier_inspect
  def hier_inspect(indent=0)
    t1 = "\t"*indent
    t2 = t1+"\t"
    s = "{\n"
    keys = self.keys
    lastkey = keys.length-1
    keys.each_with_index{ |k,i|
      v = self[k]
      s << t2
      s << k.hier_inspect(0)
      s << " => "
      s << v.hier_inspect(indent+1)
      s << "," unless i==lastkey
      s << "\n"
    }
    s << t1+"}"
  end

end

# A Hash object which raises an error if any previously-defined key attempts to be set again. 
#    foo = WriteOnceHash.new
#    foo['name'] = 'Gavin'    => 'Gavin'
#    foo['age']  = 31         => 31
#    foo['name'] = 'Bob'
#    => Error: WriteOnceHash has a value for key 'name' in object:
#       {"name"=>"Gavin", "age"=>31} (RuntimeError)
class WriteOnceHash < Hash
  # Set a value for a key; raises an error if that key already exists with a different value.
  def []=(key,val)
    raise "Error: WriteOnceHash already has a value for key '#{key.to_s}' in object:\n#{self.inspect}" if self.has_key?(key) && self[key]!=val
    super
  end
end

# Extensions to the Object class
class Object
  # Perform a deep clone on the object (where Object#clone performs a shallow clone)
  #
  # (Found on ruby-talk, forgot to include proper credit to the author; sorry)
  def deep_clone
    Marshal.load(Marshal.dump(self))
  end

  # _indent_ -- the tab indentation level to start at
  #
  # Produces similar output to #inspect, but with nice hierarchical indentation for items stored in Hashes or Arrays
  # 
  #   h = { :name=>'Gavin Kistner', 'age'=>31, :cars=>['Jeep','Passat'],
  #         "cats"=>[ {:name=>'Phleep', :weight=>15 }, {:name=>'Tessa', :weight=>7} ] }
  #
  #   puts h.inspect
  #   #=> {:cars=>["Jeep", "Passat"], :name=>"Gavin Kistner", "cats"=>[{:weight=>15, :name=>"Phleep"}, {:weight=>7, :name=>"Tessa"}], "age"=>31}
  #
  #   puts h.hier_inspect
  #   #=> {
  #   #=> 	:cars => [
  #   #=> 		"Jeep",
  #   #=> 		"Passat"
  #   #=> 	],
  #   #=> 	:name => "Gavin Kistner",
  #   #=> 	"cats" => [
  #   #=> 		{
  #   #=> 			:weight => 15,
  #   #=> 			:name => "Phleep"
  #   #=> 		},
  #   #=> 		{
  #   #=> 			:weight => 7,
  #   #=> 			:name => "Tessa"
  #   #=> 		}
  #   #=> 	],
  #   #=> 	"age" => 31
  #   #=> }
  def hier_inspect(indent=0)
    self.inspect
  end
end

# Extensions to the Struct class
class Struct
  # See Object#hier_inspect
  def hier_inspect(indent=0)
    t1 = "\t"*indent
    t2 = t1+"\t"
    s = "<struct #{self.class}\n"
    self.each_pair{ |key,val|
      s << t2
      s << (key.to_s + '=')
      s << val.hier_inspect(indent+1) << ','
      s << "\n"
    }
    s[0...-2] << "\n"+t1+">"
  end
end

# A mix of Hash and OpenStruct behavior.
#   foo = HashStruct.new
#   foo.bar = "Hello"
#   foo['jimmy'] = "World"
#   x = :jammy
#   foo[x] = "How goes it?"
#   p foo
#   #=> <HashStruct bar="Hello" jammy="How goes it?" jimmy="World">
require 'ostruct'
class HashStruct < OpenStruct
  # Access a value in the HashStruct by key
  def [](key)
    key = key.to_sym unless key.is_a?(Symbol)
    @table[key]
  end
  
  def []=(key,val)
    raise TypeError, "can't modify frozen #{self.class}", caller(1) if self.frozen?
    key = key.to_sym unless key.is_a?(Symbol)
    @table[key]=val
  end
end


# Extension to the Time class
class Time

  MONTH_NAMES = %w| January February March April May June July August September October November December |
  DAY_NAMES   = %w| Sunday Monday Tuesday Wednesday Thursday Friday Saturday |

  # Replaces every token in _formatString_ with its corresponding value. See the table below.
  #
  #    token:     description:             example:
  #    #YYYY#     4-digit year             2004
  #    #YY#       2-digit year             04
  #    #MMMM#     full month name          February
  #    #MMM#      3-letter month name      Feb
  #    #MM#       2-digit month number     02
  #    #M#        month number             2
  #    #DDDD#     full weekday name        Wednesday
  #    #DDD#      3-letter weekday name    Wed
  #    #DD#       2-digit day number       09
  #    #D#        day number               9
  #    #th#       day ordinal suffix       nd
  #    #hhh#      military/24-based hour   17
  #    #hh#       2-digit hour             05
  #    #h#        hour                     5
  #    #mm#       2-digit minute           07
  #    #m#        minute                   7
  #    #ss#       2-digit second           09
  #    #s#        second                   9
  #    #ampm#     "am" or "pm"             pm
  #    #AMPM#     "AM" or "PM"             PM
  # Non-tokens are left untouched in the output string.
  #    time = Time.parse('2/3/2004 15:37')
  #    time.custom_format('#DDDD#, #MMMM# #D##th# @ #h#:#mm##ampm#')
  #      =>  'Tuesday, February 3rd @ 3:37pm'
  #    time.custom_format('#YYYY#-#MMM#-#D#')
  #      =>  '2004-Feb-3'
  #    time.custom_format('#MM#/#DD#/#YY#')
  #      =>  '02/03/04'
  def custom_format(formatString)
    zYY   = (zYYYY=self.year.to_s)[-2..-1]
    zMM   = (zM=self.mon)<10 ? ('0'+zM.to_s) : zM.to_s
    zMMM  = (zMMMM=MONTH_NAMES[zM-1])[0...3]
    zDD   = (zD=self.day)<10 ? ('0'+zD.to_s) : zD.to_s
    zDDD  = (zDDDD=DAY_NAMES[wday])[0...3]
    zth   = (zD>=10&&zD<=20) ? 'th' : ((dMod=zD%10)==1) ? 'st' : (dMod==2) ? 'nd' : (dMod==3) ? 'rd' : 'th'
  
    zh    = (zhhh=self.hour)==0 ? 24 : zhhh; zh-=12 if (zh>12)
    zhh   = zh<10 ? ('0'+zh.to_s) : zh.to_s
    zmm   = (zm=self.min)<10 ? ('0'+zm.to_s) : zm.to_s
    zss   = (zs=self.sec)<10 ? ('0'+zs.to_s) : zs.to_s
    zAMPM = (zampm=zhhh<12 ? 'am' : 'pm').upcase

    f=formatString
    f.gsub!(/#YYYY#/,zYYYY);f.gsub!(/#YY#/,zYY);f.gsub!(/#MMMM#/,zMMMM);f.gsub!(/#MMM#/,zMMM);f.gsub!(/#MM#/,zMM.to_s);f.gsub!(/#M#/,zM.to_s);f.gsub!(/#DDDD#/,zDDDD);f.gsub!(/#DDD#/,zDDD);f.gsub!(/#DD#/,zDD);f.gsub!(/#D#/,zD.to_s);f.gsub!(/#th#/,zth);f.gsub!(/#hhh#/,zhhh.to_s);f.gsub!(/#hh#/,zhh);f.gsub!(/#h#/,zh.to_s);f.gsub!(/#mm#/,zmm);f.gsub!(/#m#/,zm.to_s);f.gsub!(/#ss#/,zss);f.gsub!(/#s#/,zs.to_s);f.gsub!(/#ampm#/,zampm);f.gsub!(/#AMPM#/,zAMPM);f
  end

end


# Extensions to the String class
class String
  # Return an xml/html safe version of the string,
  # where every &, < and > are replaced with appropriate entities
  def xmlsafe
    self.gsub(/&/,'&amp;').gsub(/</,'&lt;').gsub(/>/,'&gt;')
  end

  # Calls #xmlsafe, and then further replaces carriage returns and quote characters with entities
  def attrsafe
    xmlsafe.gsub(/[\r\n]+/,'&#13;&#10;').gsub(%r|"|,'&quot;').gsub(%r|'|,'&#39;')
  end
end



# Extension to the Kernel module
module Kernel
  # Ensures that the version number of the installed ruby is greater than or equal to the version string supplied.
  # Credit to Robert Klemme for posting this on ruby-talk.
  #
  #  ensure_version "1.7.0"
  #  ensure_version "1.8.0"
  #  ensure_version "1.8.1"
  def ensure_version( ver )
    if ( RUBY_VERSION.split(/\./).map{|x|x.to_i} <=> ver.split(/\./).map{|x|x.to_i} ) < 0
      throw "Version error: should be #{ver} but is #{RUBY_VERSION}"
    end
  end
end


# Extensions to the Numeric class
class Numeric
  # Rounds to the specified number of decimal places, returning a string value.
  #
  #   (1.234).round_to(2)   =>  '1.23'
  #   (-1.234).round_to(2)  =>  '-1.23'
  #   (-0.007).round_to(2)   => '-0.01'
  #   (-0.007).round_to(1)   => '0.0'
  def round_to(decimals)
    if self<0 then
      s='-';
      x=-self;
    else
      s='';
      x=self;
    end
    if x>=1.0e15 then
      m=x.to_s;
    else
      m=(x*10**decimals).round.to_s
      if (decimals!=0) then
        k=m.length;
        if k<=decimals then
          z='000000000000000'[0..(decimals-k)]
          m=z+m;
          k=decimals+1;
        end
        m.insert(k-decimals,'.');
      end
    end
    s='' if (/^0\.0*$/=~m);
    s+m;
  end
end


# Extensions to the Vector class
class Vector
	def normalized
		self * ( 1 / self.r )
	end
end


# Extensions to the Kernel module
module Kernel
  # Requires a file relative to the invoking file's path.
  #
  # For example, given the following directory structure:
  #   /app/run.rb
  #   /app/lib/relative_require.rb
  #   /app/code/app.rb
  #   /app/code/klass1.rb
  #   /app/code/mod1.rb
  #   /app/code/mod1/subklass.rb
  #
  # You might use code like the following:
  #
  #
  # <b>run.rb</b>
  #   require 'lib/relative_require.rb'
  #   require 'code/app.rb'
  #   App.instance.run
  #
  #
  # <b>app.rb</b>
  #   require_relative 'klass1.rb'
  #   require_relative 'mod1.rb'
  #   ...
  #
  #
  # <b>mod1.rb</b>
  #   require_relative 'mod1/subklass.rb'
  #   ...
  def require_relative( *paths )
    path_match = Regexp.new( "^.+#{File::SEPARATOR}" )
    paths.each{ |path|
      file_path = caller[2] && caller[2].match( path_match )
      full_path = file_path && file_path[0] || ''
      require full_path + path
    }
  end
end
